/*
 * KernelMAP Visualization program
 *
 * Author: j00ru//vx
 * Date:   2010-01-02
 * E-mail: j00ru.vx@gmail.com
 * WWW:    http://j00ru.vexillium.org/
 *
 * Module: CGdtLdtIdt.cpp 
 *         Retrieves information about some processor-specific structures,
 *         such as Global Descriptor Table (per-processor), Interrupt Descriptor Table
 *         (per-processor), Local Descriptor Table (per-process).
 *
 * LICENSE
 * Permission is hereby granted to use, copy, modify, and distribute this source code, 
 * or portions hereof, documentation and executables, for any purpose, without fee, 
 * subject to the following restrictions:
 * 
 * 1. The origin of this source code must not be misrepresented.
 *
 * 2. Altered versions must be plainly marked as such and must not be misrepresented 
 *    as being the original source.
 *
 * 3. This Copyright notice may not be removed or altered from any source or altered 
 *    source distribution.
 * 
 * This software is provided "AS IS". The author does not guarantee that this program works 
 * or is bug-free. The author takes no responsibility for any damage caused by this program.
 * Use at your own risk.
 */

#include <windows.h>
#include <ddk\ntapi.h>

#include "CGlobalInformation.h"
#include "CGdtLdtIdt.h"
#include "def.h"
using namespace std;


/* --------- Some very nasty (GCC-specific) hacks here
 * Do not try this at home!
 */
extern "C" void SGDT(void*);
asm("\
  	.def	___SGDT;	.scl	2;	.type	32;	.endef\n\
    .global _SGDT\n\
    _SGDT:\n\
    .intel_syntax\n\
    mov eax, [esp+4]\n\
    sgdt [eax]\n\
    ret\n\
    .att_syntax");

extern "C" void SLDT(void*);
asm("\
  	.def	___SLDT;	.scl	2;	.type	32;	.endef\n\
    .global _SLDT\n\
    _SLDT:\n\
    .intel_syntax\n\
    mov eax, [esp+4]\n\
    sldt [eax]\n\
    ret\n\
    .att_syntax");

extern "C" void SIDT(void*);
asm("\
  	.def	___SIDT;	.scl	2;	.type	32;	.endef\n\
    .global _SIDT\n\
    _SIDT:\n\
    .intel_syntax\n\
    mov eax, [esp+4]\n\
    sidt [eax]\n\
    ret\n\
    .att_syntax");



/* Information class - collecting information about some internal Windows kernel structures
 */
VOID 
CGdtLdtIdt::GetGdt(LPVOID Gdt)
{
  SGDT(Gdt);
}

VOID
CGdtLdtIdt::GetLdt(DWORD* Ldt)
{
  LDT_ENTRY LdtEntry;

  SLDT(Ldt);
  
  if(*Ldt != 0)
  {
    GetThreadSelectorEntry(GetCurrentThread(),*Ldt,&LdtEntry);
    *Ldt = (typeof(DWORD))(LdtEntry.BaseLow | (LdtEntry.HighWord.Bytes.BaseMid << 16) | (LdtEntry.HighWord.Bytes.BaseHi << 24));
  }
  else
  {
    *Ldt = 0;
  }
}

VOID 
CGdtLdtIdt::GetIdt(LPVOID Idt)
{
  SIDT(Idt);
}


CGdtLdtIdt::CGdtLdtIdt()
{
}

// class desctructor
CGdtLdtIdt::~CGdtLdtIdt()
{
}

BOOL 
CGdtLdtIdt::Run()
{
  AddressDescriptor CurrentAddress;      
  SYSTEM_INFO SystemInfo;

#pragma pack(push, 1)
  struct
  {
    WORD   Segment;
    LPVOID BaseAddress;
  } Gdt, Idt;
#pragma pack(pop)  
  DWORD Ldt;

  LogClass.Debug("[+] CGdtLdtIdt::Run() called.\n");

  /* Clear the old entries
   */
  this->KM_Addresses.clear();

  /* Get the number of processors (or cores) the system is currently running on
   */
  GetSystemInfo(&SystemInfo);
  for( ULONG i=0;i<SystemInfo.dwNumberOfProcessors;i++ )
  {
    /* Makes sure the thread is running on a specific core
     */
    SetThreadAffinityMask(GetCurrentThread(),(1<<i));

    /* Get the per-processor GDT/IDT addresses
     */
    GetGdt(&Gdt);
    GetIdt(&Idt);
        

    /* Fill the internal structure
     */
    CurrentAddress.Type           = GDT;
    CurrentAddress.Address        = Gdt.BaseAddress;
    CurrentAddress.Size           = 0x400;        // XXX: not 100% sure about this size
    this->KM_Addresses.push_back(CurrentAddress);

    CurrentAddress.Type           = IDT;
    CurrentAddress.Address        = Idt.BaseAddress;
    CurrentAddress.Size           = 256 * 8;      // XXX: however, this one should be true
    this->KM_Addresses.push_back(CurrentAddress);
  }

  /* Set the affinity mask back to all available processors
   */
  SetThreadAffinityMask(GetCurrentThread(),(1<<(SystemInfo.dwNumberOfProcessors+1))-1);

  /* Get the per-process LDT structure address, if present
   */
  GetLdt(&Ldt);
  if(Ldt != 0)
  {
    CurrentAddress.Type           = LDT;
    CurrentAddress.Address        = (LPVOID)Ldt;
    CurrentAddress.Size           = 0xfff;         // FIXME: Some dummy value here
                                                   //        Put the Limit value from segment descriptor, here
    this->KM_Addresses.push_back(CurrentAddress);
  }

  return TRUE;
}

VOID
CGdtLdtIdt::PrintConsoleInfo()
{
  LogClass.Print(FOREGROUND_INTENSITY|FOREGROUND_GREEN|FOREGROUND_BLUE,"> SYSTEM STRUCTURES\n");
  
  SYSTEM_INFO SystemInfo;
  GetSystemInfo(&SystemInfo);

  DWORD NumberOfProcessors = SystemInfo.dwNumberOfProcessors;
  LogClass.Print(0," - Number of processors:");
  LogClass.Print(FOREGROUND_RED|FOREGROUND_GREEN,"%9d\n",NumberOfProcessors);

  for( ULONG i=0;i<NumberOfProcessors;i++ )
  {
    LogClass.Print(0," - GDT[%1d]:             ",i);
    LogClass.Print(FOREGROUND_RED|FOREGROUND_GREEN,"0x%.8x\n",this->KM_Addresses[i*2].Address);

    LogClass.Print(0," - IDT[%1d]:             ",i);
    LogClass.Print(FOREGROUND_RED|FOREGROUND_GREEN,"0x%.8x\n",this->KM_Addresses[i*2+1].Address);
  }
  
  if(this->KM_Addresses.size() & 1)
  {
    LogClass.Print(0," - LDT:                ");
    LogClass.Print(FOREGROUND_RED|FOREGROUND_GREEN,"0x%.8x\n",this->KM_Addresses.back().Address);
  }
}

