/*
 * KernelMAP Visualization program
 *
 * Author: j00ru//vx
 * Date:   2010-01-02
 * E-mail: j00ru.vx@gmail.com
 * WWW:    http://j00ru.vexillium.org/
 *
 * Module: CHandleInformation.cpp 
 *         Retrieves information about all the object handles present in the system
 *         handle table. 
 *
 * LICENSE
 * Permission is hereby granted to use, copy, modify, and distribute this source code, 
 * or portions hereof, documentation and executables, for any purpose, without fee, 
 * subject to the following restrictions:
 * 
 * 1. The origin of this source code must not be misrepresented.
 *
 * 2. Altered versions must be plainly marked as such and must not be misrepresented 
 *    as being the original source.
 *
 * 3. This Copyright notice may not be removed or altered from any source or altered 
 *    source distribution.
 * 
 * This software is provided "AS IS". The author does not guarantee that this program works 
 * or is bug-free. The author takes no responsibility for any damage caused by this program.
 * Use at your own risk.
 */
#include <set>
#include <windows.h>
#include <ddk\ntapi.h>
#include <ddk\ntddk.h>
#include <ddk\winddk.h>

#include "CGlobalInformation.h"
#include "CHandleInformation.h"
#include "def.h"
using namespace std;


/* Information class - collecting information about all the active HANDLE
 * values present in the system
 */
BOOL
CKernelHandleInformation::Run()
{
  struct
  {
    ULONG NumberOfHandles;
    SYSTEM_HANDLE_TABLE_ENTRY_INFO Handles[1];
  } *KernelHandleInfo;
  DWORD dwSize = sizeof(ULONG) + sizeof(SYSTEM_HANDLE_TABLE_ENTRY_INFO);
  NTSTATUS NtStatus;

  LogClass.Debug("[+] CKernelHandleInformation::Run() called.\n");

  /* Clear the remaining, old list entries
   */
  this->KM_Addresses.clear();      

  /* Allocate some initial piece of memory
   */
  KernelHandleInfo = (typeof(KernelHandleInfo))malloc(dwSize);
  if(KernelHandleInfo == NULL)
    return FALSE;

  /* Obtain the number of bytes required to store all the HANDLE information
   */
  NtStatus = NtQuerySystemInformation(SystemHandleInformation,
                                      KernelHandleInfo,
                                      dwSize,
                                      &dwSize);


  if(NtStatus == STATUS_INFO_LENGTH_MISMATCH)
  {
    LogClass.Debug("[+] First NtQuerySystemInformation() return value: STATUS_INFO_LENGTH_MISMATCH\n");
    LogClass.Debug("[+] Number of bytes required to store the HANDLE information: %d (0x%.8x)\n",dwSize,dwSize);

    /* Free the old buffer...
     */
    free(KernelHandleInfo);

    /* ... and allocate a new, bigger one.
     * The +0x1000 value aims to amortise the list growth between
     * two, respective NtQuerySystemInformation calls.
     */
    dwSize += 0x1000;
    KernelHandleInfo = (typeof(KernelHandleInfo))malloc(dwSize);
    if(KernelHandleInfo == NULL)
      return FALSE;
  }
  else
  {
    /* XXX: Assuming that any error code other than STATUS_INFO_LENGTH_MISMATCH,
     *      including STATUS_SUCCESS indicates function failure.
     */
    return FALSE;
  }

  /* Try to retrieve the information using a new buffer
   */
  if(!NT_SUCCESS(NtStatus=
        NtQuerySystemInformation(SystemHandleInformation,
                                 KernelHandleInfo,
                                 dwSize,
                                 NULL)))
  {
    LogClass.Debug("[-] NtQuerySystemInformation(SystemHandleInformation) failed : 0x%.8x\n",(UINT)NtStatus);
    return FALSE;
  }

  /* Parse the system structures to internal format and add them
   * to the KM_Addresses list
   */
  for( ULONG i=0;i<KernelHandleInfo->NumberOfHandles;i++ )
  {
    AddressDescriptor CurrentAddress;
     
    CurrentAddress.Type     = HandleAllocation;
    CurrentAddress.Address  = KernelHandleInfo->Handles[i].Object;
    CurrentAddress.Size     = 0;

    memcpy(&CurrentAddress.AddressInfo.KernelHandle,
           &KernelHandleInfo->Handles[i],
           sizeof(SYSTEM_HANDLE_TABLE_ENTRY_INFO));

    this->KM_Addresses.push_back(CurrentAddress);
  }

  return TRUE;
}

VOID
CKernelHandleInformation::PrintConsoleInfo()
{
  LogClass.Print(FOREGROUND_INTENSITY|FOREGROUND_GREEN|FOREGROUND_BLUE,"> HANDLES\n");
  
  set<DWORD> Pages;
  ULONG NumberOfHandles = this->KM_Addresses.size();
  for( ULONG i=0;i<NumberOfHandles;i++ )
  {
    Pages.insert((DWORD)this->KM_Addresses[i].Address & 0xfffff000);
  }

  LogClass.Print(0," - Number of handles:   ");
  LogClass.Print(FOREGROUND_GREEN,"%9d (0x%.8x)\n",NumberOfHandles,NumberOfHandles);
  LogClass.Print(0," - Number of used pages:");
  LogClass.Print(FOREGROUND_GREEN,"%9d (0x%.8x)\n\n",Pages.size(),Pages.size());
}

