/*
 * KernelMAP Visualization program
 *
 * Author: j00ru//vx
 * Date:   2010-01-02
 * E-mail: j00ru.vx@gmail.com
 * WWW:    http://j00ru.vexillium.org/
 *
 * Module: CLockInformation.cpp 
 *         Retrieves available information about the kernel-mode addresses, where
 *         some lock (ERESOURCE) synchronization structures are placed.
 *
 * LICENSE
 * Permission is hereby granted to use, copy, modify, and distribute this source code, 
 * or portions hereof, documentation and executables, for any purpose, without fee, 
 * subject to the following restrictions:
 * 
 * 1. The origin of this source code must not be misrepresented.
 *
 * 2. Altered versions must be plainly marked as such and must not be misrepresented 
 *    as being the original source.
 *
 * 3. This Copyright notice may not be removed or altered from any source or altered 
 *    source distribution.
 * 
 * This software is provided "AS IS". The author does not guarantee that this program works 
 * or is bug-free. The author takes no responsibility for any damage caused by this program.
 * Use at your own risk.
 */
#include <set>
#include <windows.h>
#include <ddk\ntapi.h>

#include "CGlobalInformation.h"
#include "CLockInformation.h"
#include "def.h"
using namespace std;


/* Information class - collecting information about all the locks (synchronization
 * object) in kernel memory
 */
BOOL
CKernelLocksInformation::Run()
{
  static struct
  {
    ULONG Count;
    SYSTEM_LOCK_INFORMATION Locks[1];
  } *KernelLockInfo;
  DWORD dwSize = sizeof(ULONG) + sizeof(SYSTEM_LOCK_INFORMATION);
  NTSTATUS NtStatus;

  LogClass.Debug("[+] CKernelLocksInformation::Run() called.\n");      

  /* Clear the remaining, old list entries
   */      
  this->KM_Addresses.clear();

  /* Allocate some initial piece of memory
   */
  KernelLockInfo = (typeof(KernelLockInfo))malloc(dwSize);
  if(KernelLockInfo == NULL)
    return FALSE;        

  /* Obtain the number of bytes required to store all the MODULE information
   */
  NtStatus = NtQuerySystemInformation(SystemLocksInformation,
                                      KernelLockInfo,
                                      dwSize,
                                      &dwSize);

  if(NtStatus == STATUS_INFO_LENGTH_MISMATCH)
  {
    LogClass.Debug("[+] First NtQuerySystemInformation() return value: STATUS_INFO_LENGTH_MISMATCH\n");
    LogClass.Debug("[+] Number of bytes required to store the LOCKS information: %d (0x%.8x)\n",dwSize,dwSize);

    /* Free the old buffer...
     */
    free(KernelLockInfo);

    /* ... and allocate a new, bigger one.
     * The +0x1000 value aims to amortise the list growth between
     * two, respective NtQuerySystemInformation calls.
     */
    dwSize += 0x1000;
    KernelLockInfo = (typeof(KernelLockInfo))malloc(dwSize);
    if(KernelLockInfo == NULL)
      return FALSE;
  }
  else
  {
    /* XXX: Assuming that any error code other than STATUS_INFO_LENGTH_MISMATCH,
     *      including STATUS_SUCCESS indicates function failure.
     */
    return FALSE;
  }

  /* Try to retrieve the information using a new buffer
   */
  if(!NT_SUCCESS(NtStatus=
        NtQuerySystemInformation(SystemLocksInformation,
                                 KernelLockInfo,
                                 dwSize,
                                 NULL)))
  {
    LogClass.Debug("[-] NtQuerySystemInformation(SystemLocksInformation) failed : 0x%.8x\n",(UINT)NtStatus);
    return FALSE;
  }

  for( ULONG i=0;i<KernelLockInfo->Count;i++ )
  {
    AddressDescriptor CurrentAddress;

    CurrentAddress.Type           = LockAllocation;
    CurrentAddress.Address        = KernelLockInfo->Locks[i].Address;
    CurrentAddress.Size           = 0;

    memcpy(&CurrentAddress.AddressInfo.KernelLock,
           &KernelLockInfo->Locks[i],
           sizeof(SYSTEM_LOCK_INFORMATION));

    this->KM_Addresses.push_back(CurrentAddress);
  }
 
  return TRUE;
}

VOID
CKernelLocksInformation::PrintConsoleInfo()
{
  LogClass.Print(FOREGROUND_INTENSITY|FOREGROUND_GREEN|FOREGROUND_BLUE,"> LOCKS\n");
  
  set<DWORD> Pages;
  ULONG NumberOfLocks = this->KM_Addresses.size();
  for( ULONG i=0;i<NumberOfLocks;i++ )
  {
    Pages.insert((DWORD)this->KM_Addresses[i].Address & 0xfffff000);
  }

  LogClass.Print(0," - Number of locks:     ");
  LogClass.Print(FOREGROUND_BLUE,"%9d (0x%.8x)\n",NumberOfLocks,NumberOfLocks);
  LogClass.Print(0," - Number of used pages:");
  LogClass.Print(FOREGROUND_BLUE,"%9d (0x%.8x)\n\n",Pages.size(),Pages.size());
}

