/*
 * KernelMAP Visualization program
 *
 * Author: j00ru//vx
 * Date:   2010-01-02
 * E-mail: j00ru.vx@gmail.com
 * WWW:    http://j00ru.vexillium.org/
 *
 * Module: CModuleInformation.cpp 
 *         Retrieves information about the executable images present in the kernel
 *         memory (i.e. the kernel itself, standard drivers [win32k], device drivers etc).
 *
 * LICENSE
 * Permission is hereby granted to use, copy, modify, and distribute this source code, 
 * or portions hereof, documentation and executables, for any purpose, without fee, 
 * subject to the following restrictions:
 * 
 * 1. The origin of this source code must not be misrepresented.
 *
 * 2. Altered versions must be plainly marked as such and must not be misrepresented 
 *    as being the original source.
 *
 * 3. This Copyright notice may not be removed or altered from any source or altered 
 *    source distribution.
 * 
 * This software is provided "AS IS". The author does not guarantee that this program works 
 * or is bug-free. The author takes no responsibility for any damage caused by this program.
 * Use at your own risk.
 */
#include <windows.h>
#include <psapi.h>

#include "CGlobalInformation.h"
#include "CModuleInformation.h"
#include "def.h"
using namespace std;

/* Information class - collecting information about all the active modules
 * available in the kernel memory
 */
BOOL
CKernelImageBase::Run()
{
  static struct
  {
    ULONG Count;
    SYSTEM_MODULE_INFORMATION_ENTRY Modules[1];
  } *KernelModuleInfo = NULL;
  DWORD dwSize = sizeof(ULONG) + sizeof(SYSTEM_MODULE_INFORMATION_ENTRY); // initial structure size
  NTSTATUS NtStatus;

  LogClass.Debug("[+] CKernelImageBase::Run() called.\n");

  /* Clear the remaining, old list entries
   */
  this->KM_Addresses.clear();

  /* Allocate some initial piece of memory
   */
  KernelModuleInfo = (typeof(KernelModuleInfo))malloc(dwSize);
  if(KernelModuleInfo == NULL)
    return FALSE;

  /* Obtain the number of bytes required to store all the MODULE information
   */
  NtStatus = NtQuerySystemInformation(SystemModuleInformation,
                                      KernelModuleInfo,
                                      dwSize,
                                      &dwSize);

  if(NtStatus == STATUS_INFO_LENGTH_MISMATCH)
  {
    LogClass.Debug("[+] First NtQuerySystemInformation() return value: STATUS_INFO_LENGTH_MISMATCH\n");
    LogClass.Debug("[+] Number of bytes required to store the MODULE information: %d (0x%.8x)\n",dwSize,dwSize);

    /* Free the old buffer...
     */
    free(KernelModuleInfo);

    /* ... and allocate a new, bigger one.
     * The +0x1000 value aims to amortise the list growth between
     * two, respective NtQuerySystemInformation calls.
     */
    dwSize += 0x1000;
    KernelModuleInfo = (typeof(KernelModuleInfo))malloc(dwSize);
    if(KernelModuleInfo == NULL)
      return FALSE;
  }
  else
  {
    /* XXX: Assuming that any error code other than STATUS_INFO_LENGTH_MISMATCH,
     *      including STATUS_SUCCESS indicates function failure.
     */
    return FALSE;
  }

  /* Try to retrieve the information using a new buffer
   */
  if(!NT_SUCCESS(NtStatus=
        NtQuerySystemInformation(SystemModuleInformation,
                                 KernelModuleInfo,
                                 dwSize,
                                 NULL)))
  {
    LogClass.Debug("[-] NtQuerySystemInformation(SystemModuleInformation) failed : 0x%.8x\n",(UINT)NtStatus);
    return FALSE;
  }      

  /* Fill the entire structure with some more or less useful information
   */
  for( ULONG i=0;i<KernelModuleInfo->Count;i++ )
  {
    AddressDescriptor CurrentAddress;

    CurrentAddress.Type           = ModuleImageBase;
    CurrentAddress.Address        = KernelModuleInfo->Modules[i].Base;
    CurrentAddress.Size           = KernelModuleInfo->Modules[i].Size;

    GetDeviceDriverBaseNameA(KernelModuleInfo->Modules[i].Base,
                             CurrentAddress.AddressInfo.KernelModule.FileName,
                             MAX_PATH);

    this->KM_Addresses.push_back(CurrentAddress);
  }      
  
  return TRUE;
}

VOID
CKernelImageBase::PrintConsoleInfo()
{
  LogClass.Print(FOREGROUND_INTENSITY|FOREGROUND_GREEN|FOREGROUND_BLUE,"> MODULES\n");

  ULONG NumberOfModules = this->KM_Addresses.size();
  ULONG TotalSize = 0;
  for( ULONG i=0;i<NumberOfModules;i++ )
  {
    TotalSize += this->KM_Addresses[i].Size;
  }

  LogClass.Print(0," - Number of modules:   ");
  LogClass.Print(FOREGROUND_RED,"%9d (0x%.8x)\n",NumberOfModules,NumberOfModules);
  LogClass.Print(0," - Total size:          ");
  LogClass.Print(FOREGROUND_RED,"%9d (0x%.8x)\n",TotalSize,TotalSize);
  LogClass.Print(0," - Average size:        ");
  LogClass.Print(FOREGROUND_RED,"%9d (0x%.8x)\n\n",TotalSize/NumberOfModules,TotalSize/NumberOfModules);
}

