/*
 * KernelMAP Visualization program
 *
 * Author: j00ru//vx
 * Date:   2010-01-02
 * E-mail: j00ru.vx@gmail.com
 * WWW:    http://j00ru.vexillium.org/
 *
 * LICENSE
 * Permission is hereby granted to use, copy, modify, and distribute this source code, 
 * or portions hereof, documentation and executables, for any purpose, without fee, 
 * subject to the following restrictions:
 * 
 * 1. The origin of this source code must not be misrepresented.
 *
 * 2. Altered versions must be plainly marked as such and must not be misrepresented 
 *    as being the original source.
 *
 * 3. This Copyright notice may not be removed or altered from any source or altered 
 *    source distribution.
 * 
 * This software is provided "AS IS". The author does not guarantee that this program works 
 * or is bug-free. The author takes no responsibility for any damage caused by this program.
 * Use at your own risk.
 */
#pragma once

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <math.h>
#include <vector>
#include <gl/gl.h>
#include <gl/glu.h>
#include <SDL/sdl.h>
#include <windows.h>
#undef main

#include "CGLobalInformation.h"

/* Constant window size values
 */
#define WND_WIDTH  0x400
#define WND_HEIGHT 0x200

using namespace std;

/* SDL Window class used for the visualization
 */
class CWindow
{
  private:
    /* Size of the window
     */
    UINT Width,Height;
    
    /* List of 2D points rendered on the window
     */
    struct Point2D
    {
      UINT x,y;
      BYTE r,g,b;
    };
    vector<Point2D> XYPoints;

  public:

    /* Class constructor
     */
    CWindow(BOOL FullScreen = FALSE, INT Width = 0x400, INT Height = 0x200)
    {
      INT ScreenBpp;
      INT ScreenFlag = SDL_OPENGL;
      float Ratio;      
      if(FullScreen) ScreenFlag |= SDL_FULLSCREEN;

      LogClass.Debug("[+] CWindow::CWindow(FullScreen = %d, Width = %d, Height = %d\n",
          FullScreen,Width,Height);

      /* Set some initial values
       */
      this->Width  = Width;
      this->Height = Height;
      this->XYPoints.clear();
 
      /* ----- SDL Initialization
       */
      const SDL_VideoInfo *Info;
      SDL_Surface *Surface;

      if( SDL_Init(SDL_INIT_VIDEO)<0 )
      {
        SDL_Quit();
        return;
      }
 
      Info = SDL_GetVideoInfo();
      if(Info == NULL)
      {
        SDL_Quit();
        return;
      }

      ScreenBpp = Info->vfmt->BitsPerPixel;
      SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
      SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
      SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
      SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
      SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

      Surface = SDL_SetVideoMode(
                this->Width, this->Height, 
                ScreenBpp, ScreenFlag);

      if(Surface == NULL)
      {
        SDL_Quit();
        return;
      }

      SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

      /* ----- OpenGL Initialization
       */
      Ratio = (float)this->Width / (float)this->Height;
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glOrtho(0,this->Width,this->Height,0,0,1);
      glMatrixMode(GL_MODELVIEW);
      glDisable(GL_DEPTH_TEST);
      glLoadIdentity();
      glTranslatef(0.375,0.375,0);
    }

    /* Class destructor
     */
    ~CWindow()
    {
    }

    /* Event handling function
     *
     * Currently only Escape key and SQL_QUIT signals are supported.
     */
    BOOL Events()
    {
      SDL_Event ev;

      memset(&ev, 0, sizeof(ev));
      while(SDL_PollEvent(&ev))
      {
        switch( ev.type )
        {
          case SDL_KEYDOWN:
            if(ev.key.keysym.sym == SDLK_ESCAPE)
              return FALSE;
            break;

          case SDL_QUIT:
            return FALSE;      
        }
      }

      return TRUE;
    }

    /* Sets the color of a specified pixel
     */
    BOOL SetPixel(UINT Row, UINT Column, BYTE r, BYTE g, BYTE b)
    {
      /* Some boundary check
       */
      if(Row      >= this->Width ||
         Column   >= this->Height )
        return FALSE;

      /* Push the pixel information onto the list
       */
      Point2D NewPoint = {Row, Column, r, g, b};
      this->XYPoints.push_back(NewPoint);

      return TRUE;
    }

    /* Rendering procedure
     */
    VOID ShowScene()
    {
      LogClass.Debug("[+] CWindow::ShowScene() called.\n");

      /* Cover the entire window with black color
       */
      glClearColor(0.0,0.0,0.0,0);
      glClear(GL_COLOR_BUFFER_BIT);

      /* Draw the points
       */
      glBegin(GL_POINTS);
      for( ULONG i=0;i<this->XYPoints.size();i++ )
      {
        glColor3f( (float)this->XYPoints[i].r/255.0f,
                   (float)this->XYPoints[i].g/255.0f,
                   (float)this->XYPoints[i].b/255.0f );
        glVertex2f(this->XYPoints[i].x,this->XYPoints[i].y);
      }
      glEnd();

      /* After rendering the points once, clear the XYPoints structure
       */
      this->XYPoints.clear();
      SDL_GL_SwapBuffers();       
    }
};

