/*
 * KernelMAP Visualization program
 *
 * Author: j00ru//vx
 * Date:   2010-01-02
 * E-mail: j00ru.vx@gmail.com
 * WWW:    http://j00ru.vexillium.org/
 *
 * Module: GlobalInformation.cpp 
 *         Main application module. Manages a simple AddressContainer class etc.
 *
 * NOTE:   This application makes use of the following external modules:
 *          * SDL    - http://www.libsdl.org/
 *          * libpng - http://www.libpng.org/pub/png/libpng.html
 *          * zlib   - http://www.zlib.net/
 *
 * LICENSE
 * Permission is hereby granted to use, copy, modify, and distribute this source code, 
 * or portions hereof, documentation and executables, for any purpose, without fee, 
 * subject to the following restrictions:
 * 
 * 1. The origin of this source code must not be misrepresented.
 *
 * 2. Altered versions must be plainly marked as such and must not be misrepresented 
 *    as being the original source.
 *
 * 3. This Copyright notice may not be removed or altered from any source or altered 
 *    source distribution.
 * 
 * This software is provided "AS IS". The author does not guarantee that this program works 
 * or is bug-free. The author takes no responsibility for any damage caused by this program.
 * Use at your own risk.
 */
#include <cstdio>
#include <cstdlib>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <windows.h>
#include <ddk\ntapi.h>
#include <ddk\ntddk.h>
#include <ddk\winddk.h>

#include "CGlobalInformation.h"
#include "CModuleInformation.h"
#include "CHandleInformation.h"
#include "CLockInformation.h"
#include "CGdtLdtIdt.h"

#include "def.h"
#include "CPNG.h"
#include "CWindow.h"
#include "CLog.h"

#define _VERSION_ "v0.0.1"

using namespace std;


struct _AddressConf
{
  CHAR Name[256]; // XXX: should be always enough
  struct {BYTE r,g,b;} Color;
} AddressConf[] =
{
  {"UNKNOWN",{0x50,0x50,0x50}},
  {"MODULE", {0xff,0x00,0x00}},
  {"HANDLE", {0x00,0xff,0x00}},
  {"LOCK",   {0x00,0x00,0xff}},
  {"GDT",    {0xff,0xff,0x00}},
  {"LDT",    {0x00,0xff,0xff}},
  {"IDT",    {0xff,0x00,0xff}}
};

const CHAR* AddressTypeName[] =
{
  "UNKNOWN",
  "MODULE",
  "HANDLE",
  "LOCK",
  "GDT",
  "LDT",
  "IDT"
};


/*
 * ----- Some local configuration variables
 */

/* How often are the KM addresses retrieved [in ms]
 * This value should be set to 500 [0.5s] by default.
 */
ULONG Frequency = 500;

/* Does the user want to save the visualisation files [PNG format]?
 * If so, define the directory path, the files will be therefore named
 * in a specific manner, i.e. 1.PNG, 2.PNG etc.
 */
struct _GraphLogging
{
  BOOL Used;
  CHAR Path[MAX_PATH];

  _GraphLogging()
  {
    /* Turned off on default
     */
    this->Used = FALSE;
    memset(this->Path,0,sizeof(this->Path));
  }
} GraphLogging;

/* Does the user want to log all/some of the program activity?
 */
CLog LogClass;



/* Global address container, collecting results from all external, registered modules
 */
class CAddressContainer
{
  private:
    /* A vector, containing all the obtained KM addresses.
     * No order specified.
     */
    vector<AddressDescriptor> KM_Addresses;

    /* A list of the executive modules, each responsible for
     * retrieving a separate type of kernel-mode address list.
     */
    vector<CKernelAddrModule*> Modules;

  public:

    CAddressContainer()
    {
      /* Clear the internal STL data structures
       */
      this->KM_Addresses.clear();
      this->Modules.clear();
    }

    ~CAddressContainer()
    {
      /* TODO */
    }

    BOOL AddModule(CKernelAddrModule* ExecutiveModule)
    {
      LogClass.Debug("[+] CAddressContainer::AddModule() called.\n");

      /* Push the new module on the end of the list
       */
      this->Modules.push_back(ExecutiveModule);
      return TRUE;
    }

    BOOL RemoveModule(CKernelAddrModule* ExecutiveModule)
    {
      LogClass.Debug("[+] CAddressContainer::RemoveModule called.\n");

      /* Iterate thru all the modules;
       * Erase and return, when found.
       */
      for( vector<CKernelAddrModule*>::iterator It=this->Modules.begin();
                                                It!=this->Modules.end();
                                                It++ )
      {
        if(*It == ExecutiveModule)
        {
          this->Modules.erase(It);
          return TRUE;
        }
      }
      return FALSE;
    }

    VOID Run()
    {
      /* Clear old contents
       */
      this->KM_Addresses.clear();

      LogClass.Debug("[+] CAddressContainer::Run() called. Number of modules: %d\n",this->Modules.size());

      /* Call the ->Run() routine from each module registere before
       */
      for( vector<CKernelAddrModule*>::iterator It=this->Modules.begin();
                                                It!=this->Modules.end();
                                                It++ )
      {
        /* Ignore any internal failure
         */
        if((*It)->Run() == FALSE)
          continue;

        /* 
         */
        (*It)->PrintConsoleInfo();
        
        /* Collect the results and store them inside a single list
         */
        ULONG Size = (*It)->Size();
        for( ULONG i=0;i<Size;i++ )
        {
          AddressDescriptor AddressInfo;
          if((*It)->GetAddressInfo(i,&AddressInfo) == FALSE)
            continue;

          this->KM_Addresses.push_back(AddressInfo);
        }
      }
    }

    VOID Sort()
    {
      LogClass.Debug("[+] CAddressContainer::Sort() called.\n");

      /* Sort the list, based on the less<> structure operator
       * (currently based on the .Address value - ascending)
       */
      sort(this->KM_Addresses.begin(),
           this->KM_Addresses.end(),
           less<AddressDescriptor>() );
    }

    ULONG Size()
    {
      /* Return the current list size.
       */
      return this->KM_Addresses.size();
    }

    BOOL GetAddressDescriptor(ULONG Index, AddressDescriptor* OutputAddress)
    {
      /* Retrieve a single address descriptor
       */
      if(Index >= this->KM_Addresses.size())
        return FALSE;

      memcpy(OutputAddress,&this->KM_Addresses[Index],sizeof(AddressDescriptor));
      return TRUE;
    }    
};



VOID WaitForEnter()
{
  ReadConsole(GetStdHandle(STD_INPUT_HANDLE),NULL,0,NULL,NULL);
}


/* Print the program usage information
 */
VOID ShowHelp()
{
  LogClass.Print(FOREGROUND_BLUE|FOREGROUND_INTENSITY,"Usage:\n"
                   "  KernelMap.exe [-h] [-f <frequency>] [-g <directory>] [-d]\n"
                   "\n"
                   " Options:\n"
                   "   -h : Shows this message.\n"
                   "   -f : Defines the graphical visualisation refreshing frequency [miliseconds].\n"
                   "   -g : Defines the directory, where the PNG files will be stored.\n"
                   "        NOTE: The graphical file storage is turned off by default.\n"
                   "   -d : Turns the debug info option on.\n\n");

  LogClass.Print(0,"Press ENTER to continue...\n");
  WaitForEnter();
}

/* Parse the argument list
 *
 * Return value:
 *   - Zero     : indicates that the function succeeded
 *   - Non-zero : contains the argument number, that caused the function to stop
 */
ULONG ParseArguments(UINT argc, CHAR** argv)
{
#define CHECK_ARGC(CurrentIndex,ArgumentsLeft,TotalNumber) (CurrentIndex>=TotalNumber-ArgumentsLeft)

  for( UINT i=1;i<argc; )
  {
    if(argv[i][0] != '-' || argv[i][2] != '\0')
      return i;

    switch(argv[i][1])
    {
      case 'h':
        ShowHelp();
        ExitProcess(0);

      case 'f':
        if(CHECK_ARGC(i,1,argc))
          return i;
        
        Frequency = atoi(argv[i+1]);
        if(Frequency == 0)
          return i;

        i+=2;
        break;

      case 'g':
        if(CHECK_ARGC(i,1,argc))
          return i;

        GraphLogging.Used = TRUE;
        strncpy(GraphLogging.Path,argv[i+1],sizeof(GraphLogging.Path));
        strncat(GraphLogging.Path,"\\",sizeof(GraphLogging.Path));

        i+=2;
        break;

      case 'd':
        LogClass.SetDebugged(TRUE);
        i++;
        break;

      default:
        return i;
    }
  }

  return 0;
}



int main(INT argc, CHAR** argv)
{
  CAddressContainer        AddressContainer;  // Global address container

  /* Partial address containers
   */
  CKernelImageBase         KernelImageBase;
  CKernelHandleInformation KernelHandleInfo;
  CKernelLocksInformation  KernelLocksInfo;
  CGdtLdtIdt               KernelGdtLdtIdt;

  /* Some local variables
   */
  ULONG FileCounter = 1;
  CHAR VersionBuffer[256];

  snprintf(VersionBuffer,sizeof(VersionBuffer),"-=*( Kernel MAP %s by j00ru//vx )*=-",_VERSION_);
  SetConsoleTitle(VersionBuffer);
 
  /* Handle the user-specified arguments
   */ 
  UINT ErrNo;
  if((ErrNo=ParseArguments(argc,argv))!=0)
  {
    LogClass.Print(0,"[-] Invalid argument number %u!\n",ErrNo);
    WaitForEnter();
    return 1;
  }

  LogClass.Print(0,"(For help, use the -h program parameter.)");  

  /* Add all currently supported modules
   */
  AddressContainer.AddModule(&KernelImageBase);
  AddressContainer.AddModule(&KernelHandleInfo);
  AddressContainer.AddModule(&KernelLocksInfo);
  AddressContainer.AddModule(&KernelGdtLdtIdt);

  /* The WND_WIDTH and WND_HEIGHT values are defined in CWindow.h
   */
  LogClass.Debug("[+] Creating SDL window sized %d x %d\n",WND_WIDTH,WND_HEIGHT);
  CWindow Window(FALSE,WND_WIDTH,WND_HEIGHT);
  SDL_WM_SetCaption(VersionBuffer,NULL);  

  /* Create the specified destination directory, if needed
   */
  if(GraphLogging.Used)
  {
    if(!CreateDirectory(GraphLogging.Path,NULL) && GetLastError() != ERROR_ALREADY_EXISTS)
    {
      LogClass.Print(0,"[-] Unable to create the %s\\ directory.",GraphLogging.Path);
      WaitForEnter();
      return 1;      
    }

    LogClass.Debug("[+] %s\\ directory successfully created.",GraphLogging.Path);    
  }

  /* Main window loop, refreshing the results of all loaded modules
   */
  while(Window.Events())
  {
    CPNG PngFile;
    static CHAR PngFileName[MAX_PATH];

    /* Set the proper console coords
     */
    COORD Coord = {0,4};
    SetConsoleCursorPosition(GetStdHandle(STD_OUTPUT_HANDLE),Coord);

    /* Recalculate all KM-Addresses
     */
    AddressContainer.Run();

    /* Open the desired PNG file
     */
    if(GraphLogging.Used)
    {
      /* Format the current PNG file name and open it
       */
      snprintf(PngFileName,sizeof(PngFileName),"%s\\%d.PNG",GraphLogging.Path,(INT)FileCounter);
      PngFile.OpenFile(PngFileName,WND_WIDTH,WND_HEIGHT);
    }

    /* Iterate thru all the address descriptors
     */
    for( ULONG i=0;i<AddressContainer.Size();i++ )
    {
      AddressDescriptor AddressInfo;

      /* Get the current one
       */
      AddressContainer.GetAddressDescriptor(i,&AddressInfo);

      /* Visualise the Address+Size area on both PNG file and SDL window
       */
      for( ULONG j=(ULONG)AddressInfo.Address;j<=((ULONG)AddressInfo.Address + AddressInfo.Size);j+=0x1000 )
      {
        if(GraphLogging.Used)
        {
          PngFile.SetPixel(((j-0x80000000)/0x1000) / 0x400,
                           ((j-0x80000000)/0x1000) % 0x400,
                           AddressConf[AddressInfo.Type].Color.r,
                           AddressConf[AddressInfo.Type].Color.g,
                           AddressConf[AddressInfo.Type].Color.b);
        }

        Window.SetPixel(((j-0x80000000)/0x1000) % 0x400,
                        ((j-0x80000000)/0x1000) / 0x400,
                        AddressConf[AddressInfo.Type].Color.r,
                        AddressConf[AddressInfo.Type].Color.g,
                        AddressConf[AddressInfo.Type].Color.b);
      }
    }
    
    /* Save the graphic file, if needed
     */
    if(GraphLogging.Used)
      PngFile.Save();

    /* Render the window pixels
     */
    Window.ShowScene();

    /* Sleep for how long the user wants [default: 500ms]
     */
    Sleep(Frequency);

    FileCounter++;
  }

  /* Clear AddressContainer a little bit
   */
  AddressContainer.RemoveModule(&KernelImageBase);
  AddressContainer.RemoveModule(&KernelHandleInfo);
  AddressContainer.RemoveModule(&KernelLocksInfo); 
  AddressContainer.RemoveModule(&KernelGdtLdtIdt);

  /* Destroy the window
   */
  SDL_Quit();
  return 0;
}

